/*
    Copyright (C) 2010  (Hagen Buchwald)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package Prog1Tools;

import Prog1Tools.*;

/**
 * Plotter auf Basis der Klasse GraphicScreen zur Darstellung von Funktionsgraphen.
 *
 * @version 1.0
 * @author hagen.buchwald@kit.edu
 * @date 2010-09-06
 */
@SuppressWarnings("serial")
public class FunctionPlotter extends GraphicScreen
                             implements FunctionPlotterSpec
{
    private final int N_POINTS = 100;
    private static FunctionPlotter instance;
    
    private double leftX;
    private double rightX;
    private double lowerY;
    private double upperY;
    private double xRatio;
    private double yRatio;
    
    
    /**
     * Konstruktor.
     *
     * @param width Breite des FunctionPlotter-Fensters.
     * @param height Hoehe des FunctionPlotter-Fensters.
     */
    public FunctionPlotter (int width, int height)
    {
        super(width, height);
        setTitle("EJE FunctionPlotter");
        setOrigin(getMaxX() / 2, getMaxY() / 2);
        setXRange(-10, 10);
        setYRange(-5, 5);
    }
    
    /**
     * Statische Methode zur einfachen Erzeugung eines Funktions-Plotters.
     * Wiederholte Aufrufe dieser Methode liefern stets die gleiche Instanz (Singleton-Pattern).
     * 
     * @return singulaere Instanz der Klasse FunctionPlotter
     */
    public synchronized static FunctionPlotter getInstance ()
    {
        FunctionPlotter result = null;
        if (instance == null)
        {
            result = new FunctionPlotter(GraphicScreen.WIDTH, GraphicScreen.HEIGHT);
        }
        else
        {
            result = instance;
        }
        return result;
    }
    
    /**
     * Statische Methode zur einfachen Erzeugung eines Funktions-Plotters.
     * Wiederholte Aufrufe dieser Methode liefern stets die gleiche Instanz (Singleton-Pattern).
     * 
     * @param width Breite des FunctionPlotter-Bildschirms in Pixel
     * @param height Hoehe des FunctionPlotter-Bildschirms in Pixel
     * @return singulaere Instanz der Klasse FunctionPlotter
     */
    public synchronized static FunctionPlotter getInstance (int width, int height)
    {
        FunctionPlotter result = null;
        if (instance == null)
        {
            result = new FunctionPlotter(width, height);
        }
        else
        {
            result = instance;
        }
        return result;
    }
    
    public void setXRange (double leftX, double rightX)
    {
        this.leftX = leftX;
        this.rightX = rightX;
        xRatio = (1.0 * getMaxX() - 2 * MARGIN) / (rightX - leftX);
    }
    
    public void setYRange (double lowerY, double upperY)
    {
        this.lowerY = lowerY;
        this.upperY = upperY;
        yRatio = (1.0 * getMaxY() - 2 * MARGIN) / (upperY - lowerY);
    }
    
    public double getLeftX ()
    {
        double result;
        result = leftX;
        return result;
    }
    
    public double getRightX ()
    {
        double result;
        result = rightX;
        return result;
    }
    
    public double getLowerY ()
    {
        double result;
        result = lowerY;
        return result;
    }
    
    public double getUpperY ()
    {
        double result;
        result = upperY;
        return result;
    }
    
    public void drawAbscissae ()
    {
        int xStart = rel2absX(leftX);
        int xEnd   = rel2absX(rightX);
        drawLine(xStart, 0, xEnd, 0);
        drawLine(xStart, 0 + Y_OFFSET, xStart, 0 - Y_OFFSET);
        drawLine(xEnd, 0 + Y_OFFSET, xEnd, 0 - Y_OFFSET);
        drawText(xStart + X_OFFSET, 0 - 5 * Y_OFFSET, "" + getLeftX());
        drawText(0 + X_OFFSET, 0 - 3 * Y_OFFSET, "" + 0);
        drawText(xEnd + X_OFFSET, 0 - 5 * Y_OFFSET, "" + getRightX());
    }
    
    public void drawOrdinate ()
    {
        int yStart = rel2absY(lowerY);
        int yEnd   = rel2absY(upperY);
        drawLine(0, yStart, 0, yEnd);
        drawLine(0 - X_OFFSET, yStart, 0 + X_OFFSET, yStart);
        drawLine(0 - X_OFFSET, yEnd, 0 + X_OFFSET, yEnd);
        drawText(0 + 5 * X_OFFSET, yStart - Y_OFFSET, "" + getLowerY());
        drawText(0 + 3 * X_OFFSET, 0 - Y_OFFSET, "" + 0);
        drawText(0 + 5 * X_OFFSET, yEnd - Y_OFFSET, "" + getUpperY());
    }
    
    public void drawCoordinateSystem ()
    {
        drawAbscissae();
        drawOrdinate();
    }
    
    public void drawGraph (double[] xValues, double[] yValues, boolean isFilled)
    {
        int nPoints = xValues.length;
        int[] xPoints = new int[nPoints + 2];
        int[] yPoints = new int[nPoints + 2];
        
        for (int i = 0; i < nPoints; i++)
        {
            xPoints[i] = rel2absX(xValues[i]);
            yPoints[i] = rel2absY(yValues[i]);
        }
        
        if (!isFilled)
        {
            drawPolyline(xPoints, yPoints, nPoints);
        }
        else
        {
            xPoints[nPoints] = xPoints[nPoints - 1];
            yPoints[nPoints] = 0;
            
            xPoints[nPoints + 1] = xPoints[0];
            yPoints[nPoints + 1] = 0;
            
            drawPolygon(xPoints, yPoints, nPoints + 2, isFilled);
        }
    }
    
    public void drawGraph (FunctionSpec function, boolean isFilled)
    {
        int nPoints = N_POINTS;
        int[] xPoints = new int[nPoints + 2];
        int[] yPoints = new int[nPoints + 2];
        double x = 0.0;
        double delta = (getRightX() - getLeftX()) / (1.0 * nPoints - 1);
        
        x = getLeftX();
        for (int i = 0; i < nPoints; i++, x = x + delta)
        {
            xPoints[i] = rel2absX(x);
            yPoints[i] = rel2absY(function.f(x));
        }
        
        if (!isFilled)
        {
            drawPolyline(xPoints, yPoints, nPoints);
        }
        else
        {
            xPoints[nPoints] = xPoints[nPoints - 1];
            yPoints[nPoints] = 0;
            
            xPoints[nPoints + 1] = xPoints[0];
            yPoints[nPoints + 1] = 0;
            
            drawPolygon(xPoints, yPoints, nPoints + 2, isFilled);
        }
    }
    
    public int rel2absX (double xRelative)
    {
        int result = 0;
        result = (int) (xRelative * ratio());
        return result;
    }
    
    public int rel2absY (double yRelative)
    {
        int result = 0;
        result = (int) ((-1.0) * yRelative * ratio());
        return result;
    }
    
    public double ratio ()
    {
        double result = 1.0;
        if (xRatio < yRatio)
        {
            result = xRatio;
        }
        else
        {
            result = yRatio;
        }
        return result;
    }
    
}
    
