/*
    Copyright (C) 2010  (Hagen Buchwald)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package Prog1Tools;

import java.awt.*;
import javax.swing.*;
import Prog1Tools.Screen;

/** 
 * Grafikbasierter Farbbildschirm mit einer Standard-Aufloesung von 600 Zeilen und 800 Spalten.
 * Ziel ist es, Programmieranfaengern in Java das Erstellen visuell ansprechender Programme
 * zu ermoeglichen, wie sie in klassischen prozeduralen Programmiersprachen wie z.B. Turbo
 * Pascal ohne objektorientiertes Know-How moeglich waren.
 * 
 * @author  hagen.buchwald@kit.edu
 * @version 1.11
 * @date 2011-04-21
 */
@SuppressWarnings("serial")
public class GraphicScreen extends Screen
                           implements GraphicScreenSpec
{
    /** Anzahl der Pixel in x-Richtung der Standard-Aufloesung. */
    public static final int WIDTH = 800;
    
    /** Anzahl der Pixel in y-Richtung der Standard-Aufloesung. */
    public static final int HEIGHT = 600;
    
    /** Anzahl der Grafik-Elemente, die maximal angezeigt werden koennen. */
    public static final int ITEMS = 100000;
    
    /** x-Koordinate des Standard-Nullpunktes. */
    public static final int ORIGIN_X = 0;
    
    /** y-Koordinate des Standard-Nullpunktes. */
    public static final int ORIGIN_Y = 0;
    
    /** instance (ein Singleton) wird beim erstmaligen Gebrauch (nicht beim Laden) der Klasse erzeugt */
    private static GraphicScreen instance;
    
    private Container contentPane;
    private Graphics graphics;
    private GraphicItem[] items;
    private int numberOfItems;
    private Point origin;
    private Color foregroundColor;
    
    /**
     * Statische Methode zur einfachen Erzeugung eines graphischen Farbbildschirms.
     * Wiederholte Aufrufe dieser Methode liefern stets die gleiche Intanz (Singleton-Pattern).
     * 
     * @return singulaere Instanz der Klasse GraphicScreen
     */
    public synchronized static GraphicScreen getInstance()
    {
        int widthOffset = 0;
        int heightOffset = 0;
        String osName = "";
        boolean isWindows = false;
        boolean isMac = false;
        
        // Betriebssystem ermitteln
        try
        {
            osName = UIManager.getSystemLookAndFeelClassName();
            if (osName.contains("apple"))
            {
                isWindows = false;
                isMac = true;
            }
            else if (osName.contains("windows"))
            {
                isWindows = true;
                isMac = false;
            }
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
        
        if (isWindows)
        {
            widthOffset = 8;
            heightOffset = 27;
        }
        else if (isMac)
        {
            widthOffset = 10;
            heightOffset = 30;
        }
        
        
        if (instance == null)
        {
            instance = new GraphicScreen(WIDTH + widthOffset, HEIGHT + heightOffset);
        }
        return instance;
    }
    
    
    /**
     * Statische Methode zur individuellen Erzeugung eines graphischen Farbbildschirms.
     * Wiederholte Aufrufe dieser Methode liefern stets die gleiche Intanz (Singleton-Pattern).
     * 
     * @param width Breite des Grafikbildschirms in Pixel
     * @param height Hoehe des Grafikbildschirms in Pixel
     * @return singulaere Instanz der Klasse GraphicScreen
     */
    public synchronized static GraphicScreen getInstance(int width, int height)
    {
        if (instance == null)
        {
            instance = new GraphicScreen(width, height);
        }
        return instance;
    }
    
    
    /**
     * Dieser Konstruktor ist protected, d.h. die Klasse darf nicht von aussen instanziiert werden (siehe Singleton-Pattern).
     *
     * @param rows Anzahl der Text-Zeilen des zu erzeugenden graphischen Farbbildschirms.
     * @param columns Anzahl der Text-Spalten des zu erzeugenden graphischen Farbbildschirms.
     */
    protected GraphicScreen (int width, int height)
    {
        setDefaultCloseOperation (EXIT_ON_CLOSE);
        setSize(width, height);
        setTitle("EJE GraphicScreen");
        setVisible(true);
        setResizable(false);
        setBackground(Screen.GREY);
        contentPane = getContentPane();
        contentPane.setBackground(Screen.GREY);
        contentPane.setForeground(Screen.BLACK);
        graphics = contentPane.getGraphics();
        items = new GraphicItem[ITEMS];
        numberOfItems = 0;
        origin = new Point(ORIGIN_X, ORIGIN_Y);
        graphics.translate(ORIGIN_X, ORIGIN_Y);
    }
    
    public void clearScreen (boolean fastMode)
    {
        for (int i = 0; i < numberOfItems; i++)
        {
            items[i] = null;
        }
        numberOfItems = 0;
        repaint();
    }
    
    public void paint (Graphics g)
    {
        super.paint(g);
        for (int i = 0; i < numberOfItems; i++)
        {
            if (items[i].isVisible())
            {
            	items[i].paint(graphics);
            }
        }
    }
    
    public Color getColor ()
    {
        Color result = null;
        result = graphics.getColor();
        return result;
    }
    
    public void setColor (Color color)
    {
        this.foregroundColor = color;
        Runnable setForegroundColor = new Runnable()
        {
            public void run()
            {
                graphics.setColor(foregroundColor);
            }
        };
        try
        {
            SwingUtilities.invokeAndWait(setForegroundColor);
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
    }
    
    public void setOrigin (int x, int y)
    {
        // vorherige Verschiebung  sofern vorhanden - rueckgaengig machen
        graphics.translate( (-1) * (int) origin.getX(), (-1) * (int) origin.getY());
        
        // neue Verschiebung durchfuehren
        origin = new Point(x, y);
        graphics.translate(x, y);
    }
    
    public Point getOrigin ()
    {
        Point result;
        result = origin;
        return result;
    }
    
    public void drawOrigin ()
    {
        int x = 0;
        int y = 0;
        int length = 20;
        int distance = 10;
        int radiusOuter = 20;
        int radiusInner = 5;
        boolean isFilled = false;
        Color oldColor = null;
        
        oldColor = getColor();
        
        setColor(Screen.BLACK);
        drawLine(x, y - distance - length, x, y - distance);
        drawLine(x, y + distance + length, x, y + distance);
        drawLine(x - distance - length, y,  x - distance, y);
        drawLine(x + distance + length, y,  x + distance, y);
        drawCircle(x, y, radiusOuter, isFilled);
        setColor(Screen.RED);
        isFilled = true;
        drawCircle(x, y, radiusInner, isFilled);
        setColor(Screen.BLACK);
        isFilled = false;
        drawCircle(x, y, radiusInner, isFilled);
        
        setColor(oldColor);
        
    }
    
    public void drawLine (int x1, int y1, int x2, int y2)
    {
        Color color = graphics.getColor();
        GraphicLine l = new GraphicLine(x1, y1, x2, y2, color);
        
        graphics.drawLine(x1, y1, x2, y2);
        
        items[numberOfItems] = l;
        numberOfItems = numberOfItems + 1;
    }
    
    public void drawPolyline (int[] xPoints, int[] yPoints, int nPoints)
    {
        Color color = graphics.getColor();
        GraphicPolyline p = new GraphicPolyline(xPoints, yPoints, nPoints, color);
        
        graphics.drawPolyline(xPoints, yPoints, nPoints);
        
        items[numberOfItems] = p;
        numberOfItems = numberOfItems + 1;
    }
    
    public void drawPolygon (int[] xPoints, int[] yPoints, int nPoints, boolean isFilled)
    {
        Color color = graphics.getColor();
        GraphicPolygon p = new GraphicPolygon(xPoints, yPoints, nPoints, color, isFilled);
        
        if (!isFilled)
        {
            graphics.drawPolygon(xPoints, yPoints, nPoints);
        }
        else
        {
            graphics.fillPolygon(xPoints, yPoints, nPoints);
        }
        
        items[numberOfItems] = p;
        numberOfItems = numberOfItems + 1;
    }
    
    public void drawText (int x, int y, String text)
    {
        Color color = graphics.getColor();
        GraphicString gs = new GraphicString(x, y, text, color);
        
        graphics.drawString(text, x, y);
        
        items[numberOfItems] = gs;
        numberOfItems = numberOfItems + 1;
    }
    
    public void drawCircle (int x, int y, int radius, boolean isFilled)
    {
        Color color = graphics.getColor();
        int xCenter = x - radius;
        int yCenter = y - radius;
        int width = radius * 2;
        int height = width;
        
        GraphicOval go = new GraphicOval(xCenter, yCenter, width, height, color, isFilled);
        
        if (!isFilled)
        {
            graphics.drawOval(xCenter, yCenter, width, height);
        }
        else
        {
            graphics.fillOval(xCenter, yCenter, width, height);
        }
        
        items[numberOfItems] = go;
        numberOfItems = numberOfItems + 1;
    }
    
    public void drawOval (int x, int y, int width, int height, boolean isFilled)
    {
        Color color = graphics.getColor();
        GraphicOval go = new GraphicOval(x, y, width, height, color, isFilled);
        
        if (!isFilled)
        {
            graphics.drawOval(x, y, width, height);
        }
        else
        {
            graphics.fillOval(x, y, width, height);
        }
        
        items[numberOfItems] = go;
        numberOfItems = numberOfItems + 1;
    }
    
    public void drawRectangle (int x, int y, int width, int height, boolean isFilled)
    {
        Color color = graphics.getColor();
        GraphicRectangle gr = new GraphicRectangle(x, y, width, height, color, isFilled);
        
        if (!isFilled)
        {
            graphics.drawRect(x, y, width, height);
        }
        else
        {
            graphics.fillRect(x, y, width, height);
        }
        
        items[numberOfItems] = gr;
        numberOfItems = numberOfItems + 1;
    }
    
    public void drawRectangle3D (int x, int y, int width, int height, boolean isFilled, boolean isRaised)
    {
        Color color = graphics.getColor();
        GraphicRectangle3D gr3D = new GraphicRectangle3D(x, y, width, height, color, isFilled, isRaised);
        
        if (!isFilled)
        {
            graphics.draw3DRect(x, y, width, height, isRaised);
        }
        else
        {
            graphics.fill3DRect(x, y, width, height, isRaised);
        }
        
        items[numberOfItems] = gr3D;
        numberOfItems = numberOfItems + 1;
    }
    
    public void drawRectangleRounded (int x, int y, int width, int height, int arcWidth, int arcHeight, boolean isFilled)
    {
        Color color = graphics.getColor();
        GraphicRectangleRounded gr = new GraphicRectangleRounded(x, y, width, height, arcWidth, arcHeight, color, isFilled);
        
        if (!isFilled)
        {
            graphics.drawRoundRect(x, y, width, height, arcWidth, arcHeight);
        }
        else
        {
            graphics.fillRoundRect(x, y, width, height, arcWidth, arcHeight);
        }
        
        items[numberOfItems] = gr;
        numberOfItems = numberOfItems + 1;
    }
    
    public void drawArc (int x, int y, int width, int height, int startAngle, int arcAngle, boolean isFilled)
    {
        Color color = graphics.getColor();
        GraphicArc go = new GraphicArc(x, y, width, height, startAngle, arcAngle, color, isFilled);
        
        if (!isFilled)
        {
            graphics.drawArc(x, y, width, height, startAngle, arcAngle);
        }
        else
        {
            graphics.fillArc(x, y, width, height, startAngle, arcAngle);
        }
        
        items[numberOfItems] = go;
        numberOfItems = numberOfItems + 1;
    }
    
    public int getNumberOfItems ()
    {
        int result = 0;
        result = numberOfItems;
        return result;
    }
    
    public int getMaxX ()
    {
        int result = 0;
        result = contentPane.getWidth() - 1;
        // System.err.println("contentPane : getWidth() = " + contentPane.getWidth());
        return result;
    }
    
    public int getMaxY ()
    {
        int result = 0;
        result = contentPane.getHeight() - 1;
        // System.err.println("contentPane : getHeight() = " + contentPane.getHeight());
        return result;
    }
    
    public int getMaxNumberOfItems ()
    {
        int result = 0;
        result = ITEMS;
        return result;
    }

    public GraphicItem getItem (int index)
    {
    	GraphicItem result = null;
    	result = items[index];
    	return result;
    }
}
    
