/*
    Copyright (C) 2010  (Hagen Buchwald)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package Prog1Tools;

import java.awt.*;
import java.awt.event.*;
import javax.sound.midi.*;
import javax.swing.*;
import javax.swing.filechooser.*;
import java.io.*;
import java.util.*;
import Prog1Tools.Screen;

/** 
 * Textbasierter Farbbildschirm mit einer Standard-Aufloesung von 25 Zeilen und 80 Spalten.
 * Ziel ist es, Programmieranfaengern in Java das Erstellen visuell ansprechender Programme
 * zu ermoeglichen, wie sie in klassischen prozeduralen Programmiersprachen wie z.B. Turbo
 * Pascal ohne objektorientiertes Know-How moeglich waren.
 * 
 * @author  hagen.buchwald@kit.edu
 * @version 1.10
 * @date 2010-08-30
 */
@SuppressWarnings("serial")
public class TextScreen extends Screen
{
    /** instance (ein Singleton) wird beim erstmaligen Gebrauch (nicht beim Laden) der Klasse erzeugt */
    private static TextScreen instance;
    
    private Container contentPane;
    private TextScreenCell [][] screenView;
    
    private final int rows;
    private final int columns;
    
    /**
     * Statische Methode zur einfachen Erzeugung eines textbasierten Farbbildschirms.
     * Wiederholte Aufrufe dieser Methode liefern stets die gleiche Intanz (Singleton-Pattern).
     * 
     * @return singulaere Instanz der Klasse Screen
     */
    public synchronized static TextScreen getInstance()
    {
        boolean rowColumnDefinition = true;
        
        if (instance == null)
        {
            instance = new TextScreen(COLUMNS, ROWS, rowColumnDefinition);
        }
        return instance;
    }
    
    /**
     * Statische Methode zur individuellen Erzeugung eines textbasierten Farbbildschirms.
     * Wiederholte Aufrufe dieser Methode liefern stets die gleiche Intanz (Singleton-Pattern).
     * 
     * @param width Breite des Textbildschirms in Pixeln
     * @param height Hoehe des Textbildschirms in Pixeln
     * @return singulaere Instanz der Klasse Screen
     */
    public synchronized static TextScreen getInstance(int width, int height)
    {
        boolean rowColumnDefinition = false;
        
        if (instance == null)
        {
            instance = new TextScreen(width, height, rowColumnDefinition);
        }
        return instance;
    }
    
    /**
     * Dieser Konstruktor ist privat, d.h. die Klasse darf nicht von aussen instanziiert werden (siehe Singleton-Pattern).
     *
     * @param columnsOrWidth Anzahl der Text-Spalten bzw. Breite (in Pixeln) des zu erzeugenden Farbbildschirms.
     * @param rowsOrHeight Anzahl der Text-Zeilen bzw. Hoehe (in Pixeln) des zu erzeugenden Farbbildschirms.
     * @param rowColumDefinition bei true, werden Zeilen und Spalten angaben erwartet;
     * bei false hingegen werden Pixel-Angaben fr Breite (width bzw. columns) und Hoehe (height bzw. rows) und des Fensters erwartet
     */
    private TextScreen (int columnsOrWidth, int rowsOrHeight, boolean rowColumnDefinition)
    {
        Character beschriftung = ' ';
        Integer einer;
        choosenFile = new File(".");
        String osName = "";
        boolean isWindows = true;
        boolean isMac = false;
        int rows = 0;
        int columns = 0;
        int width = 0;
        int height = 0;
        
        // Betriebssystem ermitteln
        try
        {
            osName = UIManager.getSystemLookAndFeelClassName();
            if (osName.contains("apple"))
            {
                isWindows = false;
                isMac     = true;
                fontName  = "DialogInput";
                fontStyle = Font.PLAIN;
                fontSize  = 11;
            }
            else if (osName.contains("windows"))
            {
                isWindows = true;
                isMac     = false;
                fontName  = "DialogInput";
                fontStyle = Font.PLAIN;
                fontSize  = 14;
            }
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
        
        if (rowColumnDefinition)
        {
            this.columns = columnsOrWidth;
            this.rows = rowsOrHeight;
        }
        else
        {
            width = columnsOrWidth;
            height = rowsOrHeight;
            this.rows = ROWS;
            this.columns = COLUMNS;
        }
        
        columns = getColumns();
        rows = getRows();
              
        this.fileChooser = new JFileChooser(choosenFile);
        this.fileFilter = new FileNameExtensionFilter("CSV Dateien", "csv");
        this.selectedMouseButton = 0;
        this.selectedKeyCode = 0;
        this.selectedKeyChar = ' ';
        this.choosenColor   = Screen.WHITE;
        this.debug = false;
        
        this.selectedOption = CANCEL_OPTION;
        
        setDefaultCloseOperation (EXIT_ON_CLOSE);
        if (rowColumnDefinition)
        {
            if (isMac)
            {
                setSize(800,497);
            }
            else
            {
                setSize(727,526);
            }
        }
        else
        {
            setSize(width, height);
        }
        
        setTitle("EJE TextScreen");
        setVisible(true);
        setResizable(false);
        setBackground(Screen.GREY);
        contentPane = getContentPane();
        contentPane.setLayout(new GridLayout(rows, columns));
        contentPane.setBackground(Screen.GREY);
        contentPane.setForeground(Screen.BLACK);
        
        screenView  = new TextScreenCell [rows][columns];
        
        for (int r = 0; r < rows; r++)
        {
            for (int c = 0; c < columns; c++)
            {
                TextScreenCell cell = new TextScreenCell(r, c, beschriftung);
                cell.setBackgroundColor(Screen.GREY);
                cell.setFont(new Font(fontName, fontStyle, fontSize));
                cell.addMouseListener(mouseListener);
                contentPane.add(cell);
                screenView[r][c] = cell;
            }
        }
        
        this.addKeyListener(keyListener);
    }
    
    /**
     * Abfrage der Anzahl der Text-Zeilen des Farbbildschirms.
     * 
     * @return Anzahl der Text-Zeilen des Farbbildschirms
     */
    public int getRows ()
    {
        int result;
        result = rows;
        return result;
    }
    
    /**
     * Abfrage der Anzahl der Text-Spalten des Farbbildschirms.
     * 
     * @return Anzahl der Text-Spalten des Farbbildschirms
     */
    public int getColumns ()
    {
        int result;
        result = columns;
        return result;
    }
    
    /**
     * Abfrage der Breite des Farbbildschirms in Pixeln.
     * 
     * @return Breite des Farbbildschirms in Pixeln
     */
    public int getWidth ()
    {
        int result;
        Dimension dimension = getSize();
        result = dimension.width;
        return result;
    }
    
    /**
     * Abfrage der Hoehe des Farbbildschirms in Pixeln.
     * 
     * @return Hoehe des Farbbildschirms in Pixeln
     */
    public int getHeight ()
    {
        int result;
        Dimension dimension = getSize();
        result = dimension.height;
        return result;
    }
    
    /**
     * Abfrage des dargestellten Zeichens an einer bestimmten Position des Farbbildschirms.
     * 
     * @param row Zeilenangabe der Position
     * @param column Spaltenangabe der Position
     * @return Zeichen an der Position (row, column) des Farbbildschirms
     */
    public char getChar (int row, int column)
    {
        char result = ' ';
        boolean outOfScreen = !isValidRow(row) || !isValidColumn(column);
        
        if (outOfScreen)
        {
            // outOfScreen --> siehe Design by Contract
        }
        else
        {
            result = screenView[row][column].getChar();
        }
        
        return result;
    }
    
    /**
     * Abfrage, ob eine Zeilenangabe innerhalb der Grenzen des Farbbildschirms liegt.
     * 
     * @param row Zeilenangabe
     * @return true, wenn die Zeilenangabe innerhalb der Grenzen des Farbbildschirms liegt, false sonst
     */
    public boolean isValidRow (int row)
    {
        boolean result;
        
        result = (row >= 0) && (row < rows);
        
        return result;
    }
    
    /**
     * Abfrage, ob eine Spaltenangabe innerhalb der Grenzen des Farbbildschirms liegt.
     * 
     * @param column Spaltenangabe
     * @return true, wenn die Spaltenangabe innerhalb der Grenzen des Farbbildschirms liegt, false sonst
     */
    public boolean isValidColumn (int column)
    {
        boolean result;
        
        result = (column >= 0) && (column < columns);
        
        return result;
    }
    
    /**
     * Abfrage, welche Datei im zuletzt durch Bestaetigung geschlossenen Datei-Dialog-Fenster ausgewaehlt wurde.
     * 
     * @return Name der ausgewaehlten Datei inclusive des vollstaendigen, absoluten Pfades
     */
    public String getSelectedFile ()
    {
        String result;
        
        result = choosenFile.toString();
        
        return result;
    }
    
    /**
     * Abfrage, ob das zuletzt geschlossene Dialog-Fenster mit einer Bestaetigung oder einem Abbruch geschlossen wurde.
     */
    public int getSelectedOption ()
    {
        return selectedOption;
    }
    
    /**
     * Abfrage der Farbe, die in dem zuletzt durch Bestaetigung geschlossenen Farbauswahl-Dialog-Fenster ausgewaehlt wurde.
     * 
     * @return Farbe, die im Farbauswahl-Dialog-Fenster ausgewaehlten wurde
     */
    public Color getChoosenColor ()
    {
        Color result;
        result = choosenColor;
        return result;
    }
    
    /**
     * Abfrage der Vordergrundfarbe der Bildschirmzelle an Position (row, column).
     * 
     * @param row Zeilenangabe der Position der abzufragenden Bildschirmzelle
     * @param column Spaltenangabe der Position der abzufragenden Bildschirmzelle
     *
     * @return Vordergrundfarbe der Bildschirmzelle an Position (row, column) 
     */
    public Color getForegroundColor (int row, int column)
    {
        Color result = new Color(0, 0, 0);
        
        if (isValidRow(row) && isValidColumn(column))
        {
            result = screenView[row][column].getForegroundColor();
        }
        else
        {
            // outOfScreen --> siehe Design by Contract
        }
        
        return result;
    }
    
    /**
     * Abfrage der Hintergrundfarbe der Bildschirmzelle an Position (row, column).
     * 
     * @param row Zeilenangabe der Position der abzufragenden Bildschirmzelle
     * @param column Spaltenangabe der Position der abzufragenden Bildschirmzelle
     *
     * @return Hintergrundfarbe der Bildschirmzelle an Position (row, column) 
     */
    public Color getBackgroundColor (int row, int column)
    {
        Color result = new Color(0, 0, 0);
        
        if (isValidRow(row) && isValidColumn(column))
        {
            result = screenView[row][column].getBackgroundColor();
        }
        else
        {
            // outOfScreen --> siehe Design by Contract
        }
        
        return result;
    }
    
    /**
     * Kommando zum Setzen eines Zeichens in die Bildschirmzelle an Position (row, column).
     * 
     * @param row Zeilenangabe der Position der Bildschirmzelle
     * @param column Spaltenangabe der Position der Bildschirmzelle
     */
    private void setChar (int row, int column, char x)
    {
        boolean outOfScreen = !isValidRow(row) || !isValidColumn(column);
        
        if (outOfScreen)
        {
            // outOfScreen --> siehe Design by Contract
        }
        else
        {
            screenView[row][column].setChar(x);
        }
    }
    
    
    /**
     * Kommando zum Fuellen des gesamten Farbbildschirms mit Zahlen von 0 bis 9 (Testbild).
     * 
     */
    public void fillScreen ()
    {
        Integer einer;
        char x;
        
        for (int r = 0; r < rows; r++)
        {
            for (int c = 0; c < columns; c++)
            {
                einer = c % 10;
                x = einer.toString().charAt(0);
                setChar(r, c, x);
            }
        } 
    }
    
    /**
     * Kommando zum Fuellen des gesamten Farbbildschirms mit allen darstellbaren Zeichen (Testbild).
     * 
     */
    public void fillScreenWithVisibleUnicodeCharacters ()
    {
        int index = 33;
        char x;
        
        for (int r = 0; r < rows; r++)
        {
            for (int c = 0; c < columns; c++)
            {
                x = (char) index;
                setChar(r, c, x);
                if (index < 126)
                {
                    index = index + 1;
                }
                else if (index == 126)
                {
                    index = 161;
                }
                else if ((index > 161) && (index < 255))
                {
                    index = index + 1;
                }
                else
                {
                    index = 33;
                }
            }
        }
    }
    
    /**
     * Kommando zum Setzen der Hintergrundfarbe der Bildschirmzelle an Position (row, column).
     * 
     * @param row Zeilenangabe der Position der Bildschirmzelle
     * @param column Spaltenangabe der Position der Bildschirmzelle
     * @param color Hintergrundfarbe
     */
    public void setBackgroundColor (int row, int column, Color color)
    {
        boolean outOfScreen = !isValidRow(row) || !isValidColumn(column);
        
        if (outOfScreen)
        {
            // outOfScreen --> siehe Design by Contract
        }
        else
        {
            screenView[row][column].setBackgroundColor(color);
        }
    }
    
    
    /**
     * Kommando zum Setzen der Vordergrundfarbe der Bildschirmzelle an Position (row, column).
     * 
     * @param row Zeilenangabe der Position der Bildschirmzelle
     * @param column Spaltenangabe der Position der Bildschirmzelle
     * @param color Vordergrundfarbe
     */
    public void setForegroundColor (int row, int column, Color color)
    {
        boolean outOfScreen = !isValidRow(row) || !isValidColumn(column);
        
        if (outOfScreen)
        {
            // outOfScreen --> siehe Design by Contract
        }
        else
        {
            screenView[row][column].setForegroundColor(color);
        }
    }
    
    /**
     * Kommando zum Setzen der Vordergrundfarbe des gesamten Farbbildschirms.
     * 
     * @param color Vordergrundfarbe
     * @param fastMode falls true, wird der Farbbildschirm kurz unsichtbar, um das Einfaerben zu beschleunigen
     */
    public void setScreenForegroundColor (Color color, boolean fastMode)
    {
        if (fastMode)
        {
            this.contentPane.setVisible(false);
        }
        
        for (int r = 0; r < rows; r++)
        {
            for (int c = 0; c < columns; c++)
            {
                setForegroundColor(r, c, color);
            }
        }
        
        if (fastMode)
        {
            this.contentPane.setVisible(true);
        }
    }
    
    /**
     * Kommando zum Setzen der Vordergrundfarbe des gesamten Farbbildschirms.
     * 
     * @param color Vordergrundfarbe
     */
    public void setScreenForegroundColor (Color color)
    {
        setScreenForegroundColor(color, true);
    }
    
    /**
     * Kommando zum Zuruecksetzen der Vordergrundfarbe des gesamten Farbbildschirms.
     * Die Standard-Vordergrundfarbe ist KIT-Scharz.
     */
    public void resetScreenForegroundColor ()
    {
        setScreenForegroundColor(Screen.BLACK, true);
    }
    
    /**
     * Kommando zum Setzen der Hintergrundfarbe des gesamten Farbbildschirms.
     * 
     * @param color Hintergrundfarbe
     * @param fastMode falls true, wird der Farbbildschirm kurz unsichtbar, um das Einfaerben zu beschleunigen
     */
    public void setScreenBackgroundColor (Color color, boolean fastMode)
    {
        if (fastMode)
        {
            this.contentPane.setVisible(false);
        }
        
        for (int r = 0; r < rows; r++)
        {
            for (int c = 0; c < columns; c++)
            {
                setBackgroundColor(r, c, color);
            }
        }
        
        if (fastMode)
        {
            this.contentPane.setVisible(true);
        }
    }
    
    /**
     * Kommando zum Setzen der Hintergrundfarbe des gesamten Farbbildschirms.
     * 
     * @param color Hintergrundfarbe
     */
    public void setScreenBackgroundColor (Color color)
    {
        setScreenBackgroundColor(color, true);
    }
    
    /**
     * Kommando zum Zuruecksetzen der Hintergrund des gesamten Farbbildschirms.
     * Die Standard-Hintergrundfarbe ist KIT-Grau.
     */
    public void resetScreenBackgroundColor ()
    {
        setScreenBackgroundColor(Screen.GREY, true);
    }
    
    /**
     * Kommando zum Zuruecksetzen der Hintergrund des gesamten Farbbildschirms.
     * Die Standard-Hintergrundfarbe ist KIT-Grau.
     * 
     * @param fastMode falls true, wird der Farbbildschirm kurz unsichtbar, um den Zuruecksetzvorgang zu beschleunigen
     */
    public void resetScreenBackgroundColor (boolean fastMode)
    {
        setScreenBackgroundColor(Screen.GREY, fastMode);
    }
    
    /**
     * Kommando zum Zuruecksetzen der zur Anzeige der Zeichen verwendeten Schriftart.
     * Die Standard-Einstellungen sind plattformabhaengig.
     * 
     * @param fastMode falls true, wird der Farbbildschirm kurz unsichtbar, um den Zuruecksetzvorgang zu beschleunigen
     */
    public void resetScreenFont ()
    {
        for (int r = 0; r < rows; r++)
        {
            for (int c = 0; c < columns; c++)
            {
                screenView[r][c].setFont(new Font(this.fontName, this.fontStyle, this.fontSize));
            }
        }
    }
    
    /**
     * Kommando zum Setzen der zur Anzeige der Zeichen verwendeten Schriftart.
     * 
     * @param fontStyle Schriftstil
     * @param fontSize Schriftgroesse
     */
    public void setScreenFont (int fontStyle, int fontSize)
    {
        for (int r = 0; r < rows; r++)
        {
            for (int c = 0; c < columns; c++)
            {
                screenView[r][c].setFont(new Font(this.fontName, fontStyle, fontSize));
            }
        }
    }
    
    /**
     * Kommando zum Loeschen der Anzeige.
     * 
     * @param fastMode falls true, wird der Farbbildschirm kurz unsichtbar, um das Loeschen zu beschleunigen
     */
    public void clearScreen (boolean fastMode)
    {
        if (fastMode)
        {
            this.contentPane.setVisible(false);
        }
        char cx = ' ';
        for (int r = 0; r < rows; r++)
        {
            for (int c = 0; c < columns; c++)
            {
               setChar(r, c, cx);
            }
        }
        if (fastMode)
        {
            this.contentPane.setVisible(true);
        }
        
    }
    
    /**
     * Kommando zum Schreiben einer Zeichenfolge ab Position (row, column).
     * 
     * @param row Zeilenangabe der Position der Bildschirmzelle
     * @param column Spaltenangabe der Position der Bildschirmzelle
     * @param text zu schreibende Zeichenfolge
     */
    public void write (int row, int column, String text)
    {
        int r = row;
        int c = column;
        int index;
        char zeichen;
        boolean outOfScreen = !isValidRow(row) || !isValidColumn(column);
        
        for (index = 0; index < text.length() && !outOfScreen; index++)
        {
            zeichen = text.charAt(index);
            if (!outOfScreen)
            {
                write(r, c, zeichen);
            }
            if (isValidColumn(c + 1))
            {
                c = c + 1;
            }
            else
            {
                outOfScreen = true;
            }
        }
        
    }
    
    /**
     * Kommando zum Schreiben eines Wertes vom Typ byte ab Position (row, column).
     * 
     * @param row Zeilenangabe der Position der Bildschirmzelle
     * @param column Spaltenangabe der Position der Bildschirmzelle
     * @param number zu schreibender Wert
     */
    public void write (int row, int column, byte number)
    {
        Byte numberObject = number;
        
        write(row, column, numberObject.toString());
    }
    
    /**
     * Kommando zum Schreiben eines Wertes vom Typ short ab Position (row, column).
     * 
     * @param row Zeilenangabe der Position der Bildschirmzelle
     * @param column Spaltenangabe der Position der Bildschirmzelle
     * @param number zu schreibender Wert
     */
    public void write (int row, int column, short number)
    {
        Short numberObject = number;
        
        write(row, column, numberObject.toString());
    }
    
    /**
     * Kommando zum Schreiben eines Wertes vom Typ int ab Position (row, column).
     * 
     * @param row Zeilenangabe der Position der Bildschirmzelle
     * @param column Spaltenangabe der Position der Bildschirmzelle
     * @param number zu schreibender Wert
     */
    public void write (int row, int column, int number)
    {
        Integer numberObject = number;
        
        write(row, column, numberObject.toString());
    }
    
    /**
     * Kommando zum Schreiben eines Wertes vom Typ long ab Position (row, column).
     * 
     * @param row Zeilenangabe der Position der Bildschirmzelle
     * @param column Spaltenangabe der Position der Bildschirmzelle
     * @param number zu schreibender Wert
     */
    public void write (int row, int column, long number)
    {
        Long numberObject = number;
        
        write(row, column, numberObject.toString());
    }
    
    /**
     * Kommando zum Schreiben eines Wertes vom Typ float ab Position (row, column).
     * 
     * @param row Zeilenangabe der Position der Bildschirmzelle
     * @param column Spaltenangabe der Position der Bildschirmzelle
     * @param number zu schreibender Wert
     */
    public void write (int row, int column, float number)
    {
        Float numberObject = number;
        
        write(row, column, numberObject.toString());
    }
    
    /**
     * Kommando zum Schreiben eines Wertes vom Typ double ab Position (row, column).
     * 
     * @param row Zeilenangabe der Position der Bildschirmzelle
     * @param column Spaltenangabe der Position der Bildschirmzelle
     * @param number zu schreibender Wert
     */
    public void write (int row, int column, double number)
    {
        Double numberObject = number;
        
        write(row, column, numberObject.toString());
    }
    
    /**
     * Kommando zum Schreiben eines Zeichens an Position (row, column).
     * 
     * @param row Zeilenangabe der Position der Bildschirmzelle
     * @param column Spaltenangabe der Position der Bildschirmzelle
     * @param number zu schreibendes Zeichen
     */
    public void write (int row, int column, char character)
    {
        Character zeichen = character;
        boolean outOfScreen = !isValidRow(row) || !isValidColumn(column);
        
        if (outOfScreen)
        {
            // outOfScreen --> siehe Design by Contract
        }
        else
        {
            screenView[row][column].setText(zeichen.toString());
        }
    }
    
    /**
     * Kommando zum Schreiben eines Wertes vom Typ boolean ab Position (row, column).
     * 
     * @param row Zeilenangabe der Position der Bildschirmzelle
     * @param column Spaltenangabe der Position der Bildschirmzelle
     * @param bool zu schreibender Wert
     */
    public void write (int row, int column, boolean bool)
    {
        Boolean boolObject = bool;
        
        write(row, column, boolObject.toString());
    }
    
    /**
     * Kommando zum Zeichnen einer Linie von Position (startRow, startCol) bis Position (endRow, endCol).
     * Die Linie durch Setzen der Hintergrundfarbe der Bildschirmzellen in der Starndard-Farbe KIT-Grau gezeichnet.
     * 
     * @param startRow Zeilenangabe der Position des Linienanfangs
     * @param startCol Spaltenangabe der Position des Linienanfangs
     * @param endRow Zeilenangabe der Position des Linienendes
     * @param endCol Spaltenangabe der Position des Linienendes
     */
    public void drawLine (int startRow, int startCol, int endRow, int endCol)
    {
        drawLine(startRow, startCol, endRow, endCol, Screen.GREY2);
    }
        
    /**
     * Kommando zum Zeichnen einer Linie von Position (startRow, startCol) bis Position (endRow, endCol).
     * Die Linie durch Setzen der Hintergrundfarbe der Bildschirmzellen auf die angegebene Farbe gezeichnet.
     * 
     * @param startRow Zeilenangabe der Position des Linienanfangs
     * @param startCol Spaltenangabe der Position des Linienanfangs
     * @param endRow Zeilenangabe der Position des Linienendes
     * @param endCol Spaltenangabe der Position des Linienendes
     * @param color Hintergrundfarbe
     */
    public void drawLine (int startRow, int startCol, int endRow, int endCol, Color color)
    {
        int row;
        int col;
        int step;
        float verticalPosition;
        float horizontalPosition;
        float verticalDistance;
        float horizontalDistance;
        float verticalStepSize;
        float horizontalStepSize;
        int numberOfSteps;
        int verticalStep;
        int horizontalStep;
        boolean outOfScreen = false;
        
        outOfScreen = !isValidRow(startRow) || !isValidColumn(startCol);
        outOfScreen = outOfScreen || !isValidRow(endRow) || !isValidColumn(endCol);
        
        if (outOfScreen)
        {
            // outOfScreen --> siehe Design by Contract
        }
        else
        {
            verticalDistance = endRow - startRow;
            horizontalDistance = endCol - startCol;
            
            if (Math.abs(verticalDistance) > Math.abs(horizontalDistance))
            {
                numberOfSteps = Math.round(Math.abs(verticalDistance));
            }
            else
            {
                numberOfSteps = Math.round(Math.abs(horizontalDistance));
            }
            
            verticalStepSize = verticalDistance / numberOfSteps;
            horizontalStepSize = horizontalDistance / numberOfSteps;
            
            verticalPosition = startRow;
            horizontalPosition = startCol;
            setBackgroundColor(startRow, startCol, color);
            
            for (step = 1; step <= numberOfSteps; step++)
            {
                verticalPosition = verticalPosition + verticalStepSize;
                horizontalPosition = horizontalPosition + horizontalStepSize;
                
                row = Math.round(verticalPosition);
                col = Math.round(horizontalPosition);
                
                setBackgroundColor(row, col, color);
            }
            
            // Workaround, da trotz invokeAndWait-Konstrukt in ScreenCell
            // keine Aktualisierung des EJE-Screens erfolgt.
            // Der Aufruf von write erzwingt jedoch diesen Refresh des Screens.
            char zeichen = screenView[24][79].getChar();
            write(24, 79, ".");
            write(24, 79, zeichen);
        }
    }
    
    /**
     * Kommando zum Verschieben einer bereits vorhandenen Linie,
     * die von Position (startRow, startCol) bis Position (endRow, endCol) reicht,
     * um distance Bildschirmzellen in vertikaler Richtung,
     * d.h. nach unten oder oben, je nach Vorzeichen von distance.
     * Die Linie durch Setzen der Hintergrundfarbe der Bildschirmzellen in der Standard-Farbe KIT-Grau gezeichnet.
     * Die Farbe der Bildschirmzellen, die durch das Verschieben "freigelegt" werden,
     * wird anhand benachbarter Bildschirmzellen am Linienanfang ermittelt.
     * 
     * @param startRow Zeilenangabe der Position des Linienanfangs
     * @param startCol Spaltenangabe der Position des Linienanfangs
     * @param endRow Zeilenangabe der Position des Linienendes
     * @param endCol Spaltenangabe der Position des Linienendes
     * @param distance postive (oder negative) Anzahl von Bildschirmzellen, um die die Linie nach unten (oder oben) verschoben werden soll
     */
    public void moveLineVertical (int startRow, int startCol, int endRow, int endCol, int distance)
    {
        Color backgroundColor = screenView[startRow][startCol].getBackgroundColor();
        Color oldBackgroundColor = Screen.GREY;
        int oldStartRow = startRow;
        int oldEndRow   = endRow;
        int oldStartCol = startCol;
        int oldEndCol   = endCol;
        int step = 0;
        int increment = 0;
        boolean isVerticalLine = false;
        boolean outOfScreen;
        
        isVerticalLine = startCol == endCol;
        
        outOfScreen = !isValidRow(startRow) || !isValidColumn(startCol);
        outOfScreen = outOfScreen || !isValidRow(endRow) || !isValidColumn(endCol);
        
        if (outOfScreen)
        {
            // outOfScreen --> siehe Design by Contract
        }
        else if (distance < 0)
        {
            increment = -1;
        }
        else if (distance > 0)
        {
            increment = 1;
        }
        
        if (isValidRow(endRow + 1))
        {
            oldBackgroundColor = screenView[endRow + 1][endCol].getBackgroundColor();
        }
        else if (isValidRow(startRow - 1))
        {
            oldBackgroundColor  = screenView[startRow - 1][startCol].getBackgroundColor();
        }
        else
        {
            outOfScreen = true;
        }
        
        for (step = 1; step <= Math.abs(distance) && !outOfScreen; step++)
        {
            if ((startRow + increment) < rows)
            {
                oldStartRow = startRow;
                startRow = startRow + increment;
            }
            else
            {
                outOfScreen = true;
            }
            if ((endRow + increment) < rows)
            {
                oldEndRow = endRow;
                endRow = endRow + increment;
            }
            else
            {
                outOfScreen = true;
            }
            
            if (!outOfScreen)
            {
                if (isVerticalLine)
                {
                    if (increment == 1)
                    {
                        setBackgroundColor(oldStartRow, startCol, oldBackgroundColor);
                    }
                    if (increment == -1)
                    {
                        setBackgroundColor(oldEndRow, endCol, oldBackgroundColor);
                    }
                }
                else
                {
                    drawLine(oldStartRow, oldStartCol, oldEndRow, oldEndCol, oldBackgroundColor);
                }
                drawLine(startRow, startCol, endRow, endCol, backgroundColor);
            }
        }
    }
    
    /**
     * Kommando zum Verschieben einer bereits vorhandenen Linie,
     * die von Position (startRow, startCol) bis Position (endRow, endCol) reicht,
     * um distance Bildschirmzellen in horizontaler Richtung,
     * d.h. nach unten oder oben, je nach Vorzeichen von distance.
     * Die Linie durch Setzen der Hintergrundfarbe der Bildschirmzellen in der Standard-Farbe KIT-Grau gezeichnet.
     * Die Farbe der Bildschirmzellen, die durch das Verschieben "freigelegt" werden,
     * wird anhand benachbarter Bildschirmzellen am Linienanfang ermittelt.
     * 
     * @param startRow Zeilenangabe der Position des Linienanfangs
     * @param startCol Spaltenangabe der Position des Linienanfangs
     * @param endRow Zeilenangabe der Position des Linienendes
     * @param endCol Spaltenangabe der Position des Linienendes
     * @param distance postive (oder negative) Anzahl von Bildschirmzellen, um die die Linie nach unten (oder oben) verschoben werden soll
     */
    public void moveLineHorizontal (int startRow, int startCol, int endRow, int endCol, int distance)
    {
        Color backgroundColor = screenView[startRow][startCol].getBackgroundColor();
        Color oldBackgroundColor = Screen.GREY;
        int oldStartRow = startRow;
        int oldEndRow   = endRow;
        int oldStartCol = startCol;
        int oldEndCol   = endCol;
        int step = 0;
        int increment = 0;
        boolean isHorizontalLine = false;
        boolean outOfScreen;
        
        isHorizontalLine = startRow == endRow;
        
        outOfScreen = !isValidRow(startRow) || !isValidColumn(startCol);
        outOfScreen = outOfScreen || !isValidRow(endRow) || !isValidColumn(endCol);
        
        if (outOfScreen)
        {
            // outOfScreen --> siehe Design by Contract
        }
        else if (distance < 0)
        {
            increment = -1;
        }
        else if (distance > 0)
        {
            increment = 1;
        }
        
        if (isValidColumn(endCol + 1))
        {
            oldBackgroundColor = screenView[endRow][endCol + 1].getBackgroundColor();
        }
        else if (isValidRow(startCol - 1))
        {
            oldBackgroundColor = screenView[startRow][startCol - 1].getBackgroundColor();
        }
        else
        {
            outOfScreen = true;
        }
        
        for (step = 1; step <= Math.abs(distance) && !outOfScreen; step++)
        {
            if ((startCol + increment) < columns)
            {
                oldStartCol = startCol;
                startCol = startCol + increment;
            }
            else
            {
                outOfScreen = true;
            }
            if ((endCol + increment) < columns)
            {
                oldEndCol = endCol;
                endCol = endCol + increment;
            }
            else
            {
                outOfScreen = true;
            }
            
            if (!outOfScreen)
            {
                if (isHorizontalLine)
                {
                    if (increment == 1)
                    {
                        setBackgroundColor(startRow, oldStartCol, oldBackgroundColor);
                    }
                    if (increment == -1)
                    {
                        setBackgroundColor(endRow, oldEndCol, oldBackgroundColor);
                    }
                }
                else
                {
                    drawLine(oldStartRow, oldStartCol, oldEndRow, oldEndCol, oldBackgroundColor);
                }
                drawLine(startRow, startCol, endRow, endCol, backgroundColor);
            }
        }
    }
    
    /**
     * Kommando zum Anzeigen des Inhalts einer Zeichen-Matrix
     * auf dem Farbbildschirm ab Position (row, column)
     * 
     * @param row Zeilenangabe der Position auf dem Farbbildschirm
     * @param column Spaltenangabe der Position auf dem Farbbildschirm
     * @param matrix zwei-dimensionales Feld von Zeichen
     */
    public void toScreen (int row, int column, char[][] matrix)
    {
        boolean isValid = false;
        int matrixRows = 0;
        int matrixColumns = 0;
        int r;
        int c;
        
        if (matrix != null)
        {
            if ((matrix[0] != null) && (row + matrix.length <= rows))
            {
                if ((matrix[0] != null) && (column + matrix[0].length <= columns))
                {
                    isValid = isValidRow(row) && isValidColumn(column);
                }
            }
            
        }
        else
        {
            // !isValid --> siehe Design by Contract
        }
        
        if (isValid)
        {
            matrixRows = matrix.length;
            matrixColumns = matrix[0].length;
            
            for (r = row; r < matrixRows; r++)
            {
                for (c = column; c < matrixColumns; c++)
                {
                    write(r + row, c + column, matrix[r][c]);
                }
            }
        }
    }
    
    /**
     * Kommando eine CSV-Datei einzulesen, um einen bestimmten Bildschirm-Zustand wieder herzustellen.
     *
     * @param file einzulesende Datei
     */
    public void openScreen (File file)
    {
        FileReader fileReader;
        BufferedReader bufferedReader;
        StringTokenizer tokenizer;
        String token;
        String delimiter;
        String zeile = "";
        int index = 0;
        int row = 0;
        int column = 0;
        Character character = ' ';
        int foregroundRed   = 0;
        int foregroundGreen = 0;
        int foregroundBlue  = 0;
        int backgroundRed   = 0;
        int backgroundGreen = 0;
        int backgroundBlue  = 0;
        TextScreenCell screenCell = null;
        
        delimiter = "" + DELIMITER;
        
        try
        {
            fileReader = new FileReader(file);
            bufferedReader = new BufferedReader(fileReader);
            
            // erste Zeile der Datei enthaelt Ueberschriften
            zeile = bufferedReader.readLine();
            
            // den gesamten Screen auslesen und in die Datei schreiben
            zeile = bufferedReader.readLine();
            while (zeile != null)
            {
                tokenizer = new StringTokenizer(zeile, delimiter);
                index = 0;
                while (tokenizer.hasMoreTokens())
                {
                    token = tokenizer.nextToken();
                    switch (index)
                    {
                        case 0 : // row
                          row = Integer.parseInt(token);
                          break;
                        
                        case 1 : // column
                          column = Integer.parseInt(token);
                          break;
                        
                        case 2 : // character
                          character = token.charAt(0);
                          break;
                        
                        case 3 : // foregroundRed
                          foregroundRed = Integer.parseInt(token);
                          break;
                        
                        case 4 : // foregroundGreen
                          foregroundGreen = Integer.parseInt(token);
                          break;
                        
                        case 5 : // foregroundBlue
                          foregroundBlue = Integer.parseInt(token);
                          break;
                        
                        case 6 : // backgroundRed
                          backgroundRed = Integer.parseInt(token);
                          break;
                        
                        case 7 : // backgroundGreen
                          backgroundGreen = Integer.parseInt(token);
                          break;
                        
                        case 8 : // backgroundBlue
                          backgroundBlue = Integer.parseInt(token);
                          break;
                        
                        default: // do nothing
                    }
                    
                    index = index + 1;
                }
                
                screenCell = screenView[row][column];
                
                screenCell.setText(character.toString());
                screenCell.setForegroundColor(new Color(foregroundRed, foregroundGreen, foregroundBlue));
                screenCell.setBackgroundColor(new Color(backgroundRed, backgroundGreen, backgroundBlue));
                    
                zeile = bufferedReader.readLine();
            }
            
            bufferedReader.close();
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
    }
    
    /**
     * Kommando den aktuellen Bildschirm-Zustand in einer CSV-Datei zu speichern.
     *
     * @param file einzulesende Datei
     */
    public void saveScreen (File file)
    {
        FileWriter fileWriter;
        BufferedWriter bufferedWriter;
        String zeile = "";
        
        try
        {
            file.createNewFile();
            fileWriter = new FileWriter(file);
            bufferedWriter = new BufferedWriter(fileWriter);
            
            // erste Zeile der Datei enthaelt Ueberschriften
            zeile = zeile + "row" + DELIMITER;
            zeile = zeile + "column" + DELIMITER;
            zeile = zeile + "char" + DELIMITER;
            zeile = zeile + "foregroundRed" + DELIMITER;
            zeile = zeile + "foregroundGreen" + DELIMITER;
            zeile = zeile + "foregroundBlue" + DELIMITER;
            zeile = zeile + "backgroundRed" + DELIMITER;
            zeile = zeile + "backgroundGreen" + DELIMITER;
            zeile = zeile + "backgroundBlue";
            
            bufferedWriter.write(zeile);
            bufferedWriter.newLine();
            
            // den gesamten Screen auslesen und in die Datei schreiben
            for (int r = 0; r < rows; r++)
            {
                for (int c = 0; c < columns; c++)
                {
                    bufferedWriter.write(screenView[r][c].toString());
                    bufferedWriter.newLine();
                }
            }
            bufferedWriter.close();
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
    }
    
    /** Abfrage der durch die letzte Mausaktion angeklickten Bildschirmzelle.
     *
     * @return angeklickte Bildschirmzelle
     */
    public TextScreenCell getSelectedTextScreenCell ()
    {
        TextScreenCell result = null;
        if (debug && selectedSource != null)
        {
            System.out.println("TextScreen : getSelectedTextScreenCell = " + selectedSource);
        }
        result = (TextScreenCell) selectedSource;
        return result;
    }
    
    /** Kommando zum Loeschen einer Zeile.
     *
     * @param row Zeile, deren Inhalt vollstaendig geloescht werden soll.
     */
    public void clearLine (int row)
    {
        clearLine(row, 0);
    }
    
    /** Kommando zum Loeschen einer Zeile ab einer bestimmten Spalte.
     *
     * @param row Zeile, deren Inhalt ab Spalte col geloescht werden soll.
     * @param col Spalte, ab der der Inhalt der Zeile row geloescht werden soll.
     */
    public void clearLine (int row, int col)
    {
        clearLine(row, col, getColumns());
    }
    
    /** Kommando zum Loeschen einer Zeile ab einer bestimmten Spalte.
     *
     * @param row Zeile, deren Inhalt ab Spalte colStart bis colEnde geloescht werden soll.
     * @param colStart Spalte, ab der der Inhalt der Zeile row geloescht werden soll.
     * @param colEnde Spalte, bis zu der der der Inhalt der Zeile row geloescht werden soll.
     */
    public void clearLine (int row, int colStart, int colEnde)
    {
        int col = 0;
        if (row >= 0 && row <= getRows())
        {
            for (col = colStart; col < Math.min(colEnde, getColumns()); col = col + 1)
            {
                write (row, col, " ");
            }
        }
    }
    
    
}
